import pandas as pd
import numpy as np
import logging
from .models import CompoundingFrequency, ContributionFrequency

logger = logging.getLogger(__file__)


def compute_compound_interest(
    annual_interest_rate: float,
) -> float:
    """
    r = the interest rate (decimal)
    n = the number of times that interest is compounded per period (per month)
    t = the number of periods the money is invested for (months)

    :param annual_interest_rate:
    :param compounding_frequency:
    :return:
    """

    # annual_interest_rate / unit_t_per_year
    r = annual_interest_rate / 12
    # Number of times that interest is compounded per unit t
    n = 1
    return r / n


def compute_future_value_series(
    num_months: int,
    principal_investment: float,
    annual_interest_rate: float = 0.0,
    compounding_frequency: CompoundingFrequency | str = CompoundingFrequency.YEARLY,
    contribution_amount: float = 0.0,
    contribution_frequency: ContributionFrequency | str = ContributionFrequency.MONTHLY,
    contribution_at_start_of_compound_period: bool = True,
) -> np.ndarray:
    """Return Future Value for entire date range

    https://www.vertex42.com/Calculators/compound-interest-calculator.html#calculator
    """
    if isinstance(compounding_frequency, str):
        compounding_frequency = CompoundingFrequency(compounding_frequency)
    if isinstance(contribution_frequency, str):
        contribution_frequency = ContributionFrequency(contribution_frequency)

    num_months -= 1
    # Initialize Months
    month_range = np.arange(num_months + 1) - 1
    month_arr = (month_range % 12) + 1
    month_arr[0] = 0

    # Initialize FV array
    future_value_arr = np.zeros(num_months + 1)
    future_value_arr[0] = principal_investment

    # Initialize Input Array
    deposits_arr = np.zeros(num_months + 1)
    add_deposit_arr = np.isin(month_arr, contribution_frequency.meta.trigger_months)
    deposits_arr[add_deposit_arr] = contribution_amount

    # Initialize Interest Arr (for debugging purposes at this time)
    interest_arr = np.zeros(num_months + 1)
    accrued_interest_arr = np.zeros(num_months + 1)
    undeposited_interest = 0

    compound_interest = compute_compound_interest(
        annual_interest_rate=annual_interest_rate,
    )
    logger.debug(f"{compound_interest=}")

    current_value = principal_investment
    iter_arr = np.vstack((month_arr, deposits_arr)).T[1:]
    for i, (month, deposit) in enumerate(iter_arr):
        if contribution_at_start_of_compound_period:
            # Time to add a contribution
            current_value += deposit

        # Compound the interest
        current_interest = current_value * compound_interest
        interest_arr[i + 1] = current_interest
        undeposited_interest += current_interest
        accrued_interest_arr[i + 1] = accrued_interest_arr[i] + current_interest
        if month in compounding_frequency.meta.trigger_months:
            # Time to compound
            current_value += undeposited_interest
            undeposited_interest = 0

        if not contribution_at_start_of_compound_period:
            # Time to add a contribution
            current_value += deposit

        future_value_arr[i + 1] = current_value

    return future_value_arr



from selenium import webdriver
from selenium.webdriver.chrome.service import Service as ChromeService
from selenium.webdriver import ChromeOptions
from webdriver_manager.chrome import ChromeDriverManager
from selenium.webdriver.common.desired_capabilities import DesiredCapabilities
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.common.by import By
from selenium.webdriver.support import expected_conditions as EC
import pandas as pd
from pathlib import Path

import tests
from tests.test_utils import TestComputeFutureValueSeries

output_dir = Path(tests.__file__).parent / "data"

capabilities = DesiredCapabilities().CHROME
capabilities["pageLoadStrategy"] = "eager"  #  interactive
service = ChromeService(ChromeDriverManager().install())
options = ChromeOptions()
options.add_argument("--headless")


def download_amortization_table(link: str, output_fpath: Path):
    print("Getting driver ready")
    driver = webdriver.Chrome(
        options=options, service=service, desired_capabilities=capabilities
    )

    print("Getting link data")
    driver.get(link)

    # Wait until we can click the "monthly" -- No need to actually click it.. but we need the wait.
    mark = (By.CSS_SELECTOR, "label[for=radio-months]")
    monthly_radio_button = WebDriverWait(driver, 20).until(
        EC.element_to_be_clickable(mark)
    )

    print("Getting Table by id with wait...")
    mark = (By.CSS_SELECTOR, "#monthlyAmortization")
    table = driver.find_element(*mark)
    df = pd.read_html(table.get_attribute("outerHTML"))[0]

    print(f"Outputting data to {output_fpath}")
    df.to_csv(output_fpath, index=False)

    driver.quit()


for test_set in TestComputeFutureValueSeries.TESTS:
    link = test_set["link"]
    name = test_set["name"]
    print(f'Grabbing table for "{name}"')
    print(link)
    output_fname = "_".join(name.split()) + ".csv"
    output_fpath = output_dir / output_fname
    if output_fpath.exists():
        print(f"\tAlready exists. Skipping {output_fname}")
    else:
        download_amortization_table(link, output_fpath)



import pytest
import pandas as pd
from pathlib import Path
from retirepy.future_value import compute_future_value_series


class TestComputeFutureValueSeries:
    """
    Use https://www.thecalculatorsite.com/finance/calculators/compoundinterestcalculator.php
    for validation
    """

    TESTS = [
        {
            "name": "monthly compounding interest with no contribution",
            "link": "https://www.thecalculatorsite.com/compound?a=12000&p=7&pp=yearly&y=1&m=0&rd=0&rp=monthly&rt=deposit&rw=0&rwp=1m&rm=beginning&ci=monthly&ip=&c=1&di=",
            "input": dict(
                num_months=12 + 1,
                principal_investment=12000.0,
                annual_interest_rate=0.07,
                compounding_frequency="Monthly",
                contribution_amount=0.0,
                contribution_frequency="Monthly",
                contribution_at_start_of_compound_period=True,
            ),
            "expected": 12_867.48,
        },
        {
            "name": "monthly compounding interest with monthly contribution at start of period",
            "link": "https://www.thecalculatorsite.com/compound?a=12000&p=7&pp=yearly&y=1&m=0&rd=100&rp=monthly&rt=deposit&rw=0&rwp=1m&rm=beginning&ci=monthly&ip=&c=1&di=",
            "input": dict(
                num_months=12 + 1,
                principal_investment=12000.0,
                annual_interest_rate=0.07,
                compounding_frequency="Monthly",
                contribution_amount=100.0,
                contribution_frequency="Monthly",
                contribution_at_start_of_compound_period=True,
            ),
            "expected": 14_113.97,
        },
        {
            "name": "monthly compounding interest with monthly contribution at end of period",
            "link": "https://www.thecalculatorsite.com/compound?a=12000&p=7&pp=yearly&y=1&m=0&rd=100&rp=monthly&rt=deposit&rw=0&rwp=1m&rm=end&ci=monthly&ip=&c=1&di=",
            "input": dict(
                num_months=12 + 1,
                principal_investment=12000.0,
                annual_interest_rate=0.07,
                compounding_frequency="Monthly",
                contribution_amount=100.0,
                contribution_frequency="Monthly",
                contribution_at_start_of_compound_period=False,
            ),
            "expected": 14_106.74,
        },
        {
            "name": "yearly compounding interest with no contribution",
            "link": "https://www.thecalculatorsite.com/compound?a=12000&p=7&pp=yearly&y=10&m=0&rd=0&rp=monthly&rt=deposit&rw=0&rwp=1m&rm=beginning&ci=yearly&ip=&c=1&di=",
            "input": dict(
                num_months=12 * 10 + 1,
                principal_investment=12000.0,
                annual_interest_rate=0.07,
                compounding_frequency="Yearly",
                contribution_amount=0.0,
                contribution_frequency="Yearly",
                contribution_at_start_of_compound_period=True,
            ),
            "expected": 23_605.82,
        },
        {
            "name": "yearly compounding interest with yearly contribution at start of period",
            "link": "https://www.thecalculatorsite.com/compound?a=12000&p=7&pp=yearly&y=10&m=0&rd=100&rp=yearly&rt=deposit&rw=0&rwp=1m&rm=beginning&ci=yearly&ip=&c=1&di=",
            "input": dict(
                num_months=12 * 10 + 1,
                principal_investment=12000.0,
                annual_interest_rate=0.07,
                compounding_frequency="Yearly",
                contribution_amount=100.0,
                contribution_frequency="Yearly",
                contribution_at_start_of_compound_period=True,
            ),
            "expected": 25_084.18,
        },
        {
            "name": "yearly compounding interest with monthly contribution at start of period",
            "link": "https://www.thecalculatorsite.com/compound?a=12000&p=7&pp=yearly&y=10&m=0&rd=100&rp=monthly&rt=deposit&rw=0&rwp=1m&rm=beginning&ci=yearly&ip=&c=1&di=",
            "input": dict(
                num_months=12 * 10 + 1,
                principal_investment=12000.0,
                annual_interest_rate=0.07,
                compounding_frequency="Yearly",
                contribution_amount=100.0,
                contribution_frequency="Monthly",
                contribution_at_start_of_compound_period=True,
            ),
            "expected": 40_814.20,
        },
    ]

    @pytest.mark.parametrize(
        argnames=["input_dict", "expected_output", "check_link"],
        argvalues=[(_["input"], _["expected"], _["link"]) for _ in TESTS],
        ids=[_["name"] for _ in TESTS],
    )
    def test_future_values_final_value(
        self, request, input_dict: dict, expected_output: float, check_link: str
    ):
        fv_series = compute_future_value_series(**input_dict)
        fail_msg = f"Failed, check link: {check_link}"
        try:
            assert (fv_series[-1]).round(2) == expected_output, fail_msg
        except:
            data_dir = Path(__file__).parent / "data"
            data_fname = "_".join(request.node.callspec.id.split()) + ".csv"
            data_fpath = data_dir / data_fname
            if data_fpath.exists():
                df = pd.read_csv(data_fpath)

                print(fv_series)
                print(df["Balance"].values)

                comparison_df = pd.DataFrame(
                    data={
                        "actual_balance": fv_series,
                        "expected_balance": df["Balance"].values,
                    },
                    index=df.index,
                )
                print(comparison_df)

                # TODO: Compare expected to actual
            else:
                print("Unable to find csv file with expected data")
            raise



from enum import Enum
from pydantic import BaseModel, validator, PrivateAttr

COMPOUNDING_TRIGGER_MONTHS = {
    12: [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12],
    4: [3, 6, 9, 12],
    2: [6, 12],
    1: [12],
}

CONTRIBUTION_TRIGGER_MONTHS = {
    12: [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12],
    4: [1, 4, 7, 10],
    2: [1, 7],
    1: [1],
}

TRIGGER_MONTH_DICT = {
    "Compounding": COMPOUNDING_TRIGGER_MONTHS,
    "Contribution": CONTRIBUTION_TRIGGER_MONTHS,
}


class FrequencyType(str, Enum):
    COMPOUNDING: str = "Compounding"
    CONTRIBUTION: str = "Contribution"


class FrequencyMeta(BaseModel):
    name: str
    per_year: int
    frequency_type: FrequencyType
    _trigger_months: list[int] = PrivateAttr()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._trigger_months = TRIGGER_MONTH_DICT[self.frequency_type.value][
            self.per_year
        ]

    @validator("per_year")
    def ensure_per_year_in_allowed(cls, v):
        if v not in [1, 2, 4, 12]:
            raise ValueError(
                "per_year can only be 1 (yearly), 2 (semi-annually), 4 (quarterly), 12 (monthly)"
            )
        return v

    @property
    def per_month(self) -> float:
        return self.per_year / 12

    @property
    def months_between(self) -> int:
        return 12 // self.per_year

    @property
    def trigger_months(self) -> list[int]:
        return self._trigger_months


MONTHLY_CONTRIBUTION_META = FrequencyMeta(
    name="Monthly", per_year=12, frequency_type="Contribution"
)
YEARLY_CONTRIBUTION_META = FrequencyMeta(
    name="Yearly", per_year=1, frequency_type="Contribution"
)
MONTHLY_COMPOUNDING_META = FrequencyMeta(
    name="Monthly", per_year=12, frequency_type="Compounding"
)
YEARLY_COMPOUNDING_META = FrequencyMeta(
    name="Yearly", per_year=1, frequency_type="Compounding"
)


class ContributionFrequency(str, Enum):
    MONTHLY: str = "Monthly"
    YEARLY: str = "Yearly"

    @property
    def meta(self) -> FrequencyMeta:
        return {
            "Monthly": MONTHLY_CONTRIBUTION_META,
            "Yearly": YEARLY_CONTRIBUTION_META,
        }[self.value]


class CompoundingFrequency(str, Enum):
    MONTHLY: str = "Monthly"
    YEARLY: str = "Yearly"

    @property
    def meta(self) -> FrequencyMeta:
        return {
            "Monthly": MONTHLY_COMPOUNDING_META,
            "Yearly": YEARLY_COMPOUNDING_META,
        }[self.value]


OccurrenceFrequency = CompoundingFrequency



import numpy as np
from pydantic import BaseModel
from retirepy.models import (
    CompoundingFrequency,
    ContributionFrequency,
    OccurrenceFrequency,
)
from retirepy.future_value import compute_future_value_series


class CashFlow(BaseModel):
    name: str

    def future_value(self, num_months: int):
        raise NotImplementedError()


class Investment(CashFlow):
    principal_investment: float
    annual_interest_rate: float
    compounding_frequency: CompoundingFrequency = ContributionFrequency.YEARLY
    contribution_amount: float = 0.0
    contribution_frequency: ContributionFrequency = ContributionFrequency.MONTHLY
    contribution_at_start_of_compound_period: bool = True

    def future_value(self, num_months: int):
        return compute_future_value_series(
            num_months=num_months,
            principal_investment=self.principal_investment,
            annual_interest_rate=self.annual_interest_rate,
            compounding_frequency=self.compounding_frequency,
            contribution_amount=self.contribution_amount,
            contribution_frequency=self.contribution_frequency,
            contribution_at_start_of_compound_period=self.contribution_at_start_of_compound_period,
        )


class EarnedIncome(CashFlow):
    starting_salary: float
    annual_percentage_increase: float

    def future_value(self, num_months: int):
        return compute_future_value_series(
            num_months=num_months,
            principal_investment=self.starting_salary,
            annual_interest_rate=self.annual_percentage_increase,
            compounding_frequency=CompoundingFrequency.YEARLY,
        )


class OneTimeIncome(CashFlow):
    amount: float
    month: int

    def future_value(self, num_months: int):
        fv = np.zeros(num_months)
        fv[self.month] = self.amount
        return fv


class OneTimeExpense(CashFlow):
    amount: float
    month: int

    def future_value(self, num_months: int):
        fv = np.zeros(num_months)
        fv[self.month] = self.amount
        return fv


class RecurringExpense(CashFlow):
    amount: float
    frequency: OccurrenceFrequency

    def future_value(self, num_months: int):
        months = np.arange(0, num_months) % 12 + 1
        fv = np.zeros(num_months)
        fv[np.isin(months, self.frequency.meta.trigger_months)] = self.amount
        return fv
